#!/usr/bin/env perl
use warnings;
use strict;

use Getopt::Std;
use File::Basename;

BEGIN {
  use FindBin qw($RealBin);
  BEGIN {
    unless ($ENV{'PPH'}) {
      my $binpath = $RealBin;
      $binpath =~ s|/[^/]+$||;
      $ENV{'PPH'} = $binpath;
    }
  }
  use lib "$ENV{PPH}/perl";
}

use PPH::Config;

my %opts;
getopts('fvr:', \%opts) or die "Wrong argument(s)\n";
my $FORCE   = $opts{'f'} || 0;
my $VERBOSE = $opts{'v'} || 0;

# Check if we are running under SGE or LSF and if this is an array job
my ($grid_mode, $jobid, $taskid, $maxtask, $grid_msg);
if (exists $ENV{'SGE_ROOT'} && exists $ENV{'JOB_ID'}) {       # SGE
  $grid_mode = 'SGE';
  $jobid  = $ENV{'JOB_ID'} ne 'undefined' ? $ENV{'JOB_ID'} : 0;
  $taskid = exists $ENV{'SGE_TASK_ID'} && $ENV{'SGE_TASK_ID'} ne 'undefined' ? $ENV{'SGE_TASK_ID'} : 0;
} elsif ($ENV{'LSB_QUEUE'} && exists $ENV{'LSB_JOBID'}) {     # LSF
  $grid_mode = 'LSF';
  $jobid  = $ENV{'LSB_JOBID'};
  $taskid = exists $ENV{'LSB_JOBINDEX'} ? $ENV{'LSB_JOBINDEX'} : 0;
}
my $array_mode = $jobid && $taskid ? 1 : 0;

if ($grid_mode) {
  # Adjust PATH for a grid job
  if ($CONFIG{'GRID_PATH'}) {
    my $ARCH = `uname -m`; chomp $ARCH;
    $ENV{'PATH'} = eval "qq{$CONFIG{GRID_PATH}}";
  }
  $grid_msg = "I am $grid_mode job $jobid \@ $ENV{HOSTNAME}\n";
}

if ($array_mode) {
  # GGI_STRIPES overwrites any array size specified initially
  $maxtask = $ENV{'GGI_STRIPES'} || $ENV{'SGE_TASK_LAST'} || $ENV{'LSB_JOBINDEX_END'};
  $grid_msg = "I am $grid_mode job $jobid task $taskid of $maxtask \@ $ENV{HOSTNAME}\n";
}

my ($stripe_no, $stripes);
if ($opts{'r'}) {
  ($stripe_no, $stripes) = $opts{'r'} =~ m|^(\d+)/(\d+)$|;
  die "ERROR: Illegal option's argument: -r $opts{'r'}\n" unless $stripe_no && $stripes;
}

# Debugging
warn $grid_msg if defined $grid_msg;

if ($array_mode) {
  warn "WARNING: Explicit input stripe specification via '-r' option ignored in $grid_mode array mode\n" if $opts{'r'};
  ($stripe_no, $stripes) = ($taskid, $maxtask);
}

# Count how many query sequences we have in input
my $input_file_count = 0;
my ($range_start, $range_end);
if ($stripes) {
  foreach my $inputfile (glob join(' ', @ARGV)) { $input_file_count++; }
  # Calculate stripe range
  ($range_start, $range_end) = stripe_range($stripe_no, $stripes, $input_file_count);
  # Debug
  my $stripe_size = $range_end - $range_start + 1;
  warn "Stripe $range_start-$range_end/$input_file_count ($stripe_size):\n";
}

my $fileno = 0;
# filename glob example, double-quote for SGE: 'uc\*.aln'
foreach my $inputfile (glob join(' ', @ARGV)) {

  $fileno++;
  if ($stripes) {
    next if $fileno < $range_start;
    last if $fileno > $range_end;
  }

  my ($name, $path, $ext)   = fileparse($inputfile, qw{ .aln });
  my $psicfile = "$name.prf";
  unless ($FORCE) {
    next if -e $psicfile && -s $psicfile;
  }
  my $cmd = "$CONFIG{PSIC} $CONFIG{PSIC_ARG} $inputfile $CONFIG{MATRIX}/Blosum62.txt $psicfile 2>&1";
  my $rc  = `$cmd`;
  warn "WARNING: ($name) PSIC failure: $rc\n" if length $rc;

}

if ($stripes) {
  warn "Stripe $range_start-$range_end/$input_file_count completed.\n";
} else {
  warn "All done.\n";
}

# Calculate [start,end] range of a stripe in dataset
# Note: all indices passed and returned are one-based
#
# Arguments:  stripe index, number of stripes, dataset size
# Returns:    data start index, data end index; both indices
#             returned are zero if no more data fall within
#             a stripe specified (e.g. number of stripes is
#             larger than the number of data items)
#
sub stripe_range {
  my ($stripe_no, $stripes, $size) = @_;

  die "ERROR: Stripe index too large: $stripe_no > $stripes\n" if $stripe_no > $stripes;

  my $p = int($size / $stripes);
  my $r = $size % $stripes;

  my $shift_start = ($stripe_no <= $r ? $stripe_no : $r + 1) - 1;
  my $shift_end   = ($stripe_no <= $r ? $stripe_no : $r);

  my $start = ($stripe_no - 1)*$p + 1 + $shift_start;
  my $end   = $stripe_no*$p + $shift_end;

  if ($start > $end) {
    return (0, 0);
  } else {
    return ($start, $end);
  }
}
