#!/usr/bin/env perl
use strict;

use Getopt::Std;

my %opts;
getopts('nx:', \%opts) or die "ERROR: Wrong option(s)\n";
# Default filtering stringency=1, remove all metagenome project sequences
my $stringency = exists $opts{'x'} ? $opts{'x'} : 1;
die "ERROR: Illegal stringency option argument: $stringency" unless $stringency =~ m{^\d$};
my $nohuman    = $opts{'n'} || 0;

my ($count_all, $count_skipped);
my $skip = 0;
while (<>) {
  if (/^>/) {
    if (/^>UniRef(\d+)_(\S+)(\s+.+)/) {
      $count_all++;
      my ($uid, $acc, $definition) = ($1, $2, $3);
      if ($stringency) {
        $skip = 0;
        my $tax   = $1 if $definition =~ /\bTax=(.+?)(?:(?:\s+\S+=)|\s*$)/;
        my $repid = $1 if $definition =~ /\bRepID=(.+?)(?:(?:\s+\S+=)|\s*$)/;
        # skip metagenome and unknown/unclassified viral & bacterial sequences
        $skip = 1 if $stringency >= 1 && ($tax =~ /\bmetagenomes?\b/i || $tax =~ /\buncultured\b/i || $repid =~ /_9ZZZZ$/);
        # also skip all taxonomically unclear sequences
        $skip = 1 if $stringency >= 2 && ($tax =~ /\broot\b/ || $tax =~ /\bunknown\b/);
        # also skip all UniParc sequences
        $skip = 1 if $stringency >= 3 && $repid =~ /\bUPI\w+\b/;
        # skip all human sequences
        $skip = 1 if $nohuman && $tax =~ /\bHomo\s+sapiens\b/;
        $count_skipped++, next if $skip;
      }
      # replace square brackets in defline to avoid confusing buggy BLAST+ Sequin parser;
      # for details, see: http://www.ncbi.nlm.nih.gov/Sequin/modifiers.html
      $definition =~ tr/[]/{}/;
      $_ = ">gnl|uniref$uid|$acc$definition\n";
    } else {
      die "Unsupported defline format:\n$_";
    }
  } else {
    # Replace Selenocysteine (U) with Cysteine (C) and Pyrrolysine (O) with
    # Lysine (K) in the sequence since MAFFT as well as many other tools
    # do not recognize these non-standard residue codes.
    tr/UOuo/CKck/;
  }
  print unless $skip;
}

warn "Filter stringency: $stringency\n";
warn "Human entries removed\n" if $nohuman;
warn "Entries processed: " . sprintf("%10s", $count_all) . "\n";
if ($stringency) {
warn "Entries skipped:   " . sprintf("%10s (%.2g%%)", $count_skipped, $count_skipped / $count_all * 100) . "\n";
warn "Entries written:   " . sprintf("%10s", $count_all - $count_skipped) . "\n";
}
